local function intersection(rpx,rpy,rdx,rdy,spx,spy,sdx,sdy)
    local rmag = math.sqrt(rdx*rdx + rdy*rdy)
    local smag = math.sqrt(sdx*sdx + sdy*sdy)
    if rdx/rmag == sdx/smag and rdy/rmag == sdy/smag then return end
    local T2 = (rdx*(spy-rpy) + rdy*(rpx-spx))/(sdx*rdy - sdy*rdx);
	local T1 = (spx+sdx*T2-rpx)/rdx;
    if T1<0 then return end
	if T2<0 or T2>1 then return end
	return rpx+rdx*T1,rpy+rdy*T1,T1
end

local function clockwise(ox,oy,mx,my,tx,ty)
    return (mx-ox)*(ty-my) - (tx-mx)*(my-oy) > 0
end

local function distance(ox,oy,tx,ty)
    return math.sqrt((tx-ox)^2 + (ty-oy)^2)
end

local function polygon(verts)
    local i,c = 0,#verts/2
    return function()
       i = i + 1
        if i<= c then
            return verts[i*2-1],verts[i*2],i
        end
    end
end

local pi_2 = 2* math.pi
local function norm(rad)
    local v = rad%pi_2
    return v<0 and v+pi_2 or v
end

local function clamp(tar,min,max)
    tar,min,max = norm(tar),norm(min),norm(max)
    if max>min then
        if tar<min then return min end
        if tar>max then return max end
        return tar
    else
        if tar>min then return tar end
        if tar<max then return tar end
        if tar-max > min-tar then return min end
        return max
    end
end

local lighter = {
    polys = {},
    lights = {},
}

local function newCircleMesh(segments)
	segments = segments or 40
	local vertices = {}
	table.insert(vertices, {0, 0})
	for i=0, segments do
		local angle = (i / segments) * math.pi * 2
		local x = math.cos(angle)
		local y = math.sin(angle)
		table.insert(vertices, {x, y,_,_,_,_,_,0})
	end
	return love.graphics.newMesh(vertices, "fan")
end

function lighter:newWorld(w,h,color)
    self.canvas = love.graphics.newCanvas()
    self.worldW = w
    self.worldH = h 
    self.darkness = color or {0,0,0,1}
    self:addBox(-10,-10,w+20,h+20)
    self.circle = newCircleMesh()
end


function lighter:addLight(conf)
    local light = {
        x = conf.x or 0,
        y = conf.y or 0,
        r = conf.r or self.worldW,
        dir = conf.dir or 0,
        arc = conf.arc or 2*math.pi,
        color = conf.color or {1,1,1,1},
        grad = conf.grad,
        diff = conf.diff or math.pi/100, --衍射
        power = conf.power or 1
    }
    table.insert(self.lights,light)
    self.dirty = true
    return light
end

function lighter:addBox(x,y,w,h)
    local poly = {}
    table.insert(self.polys,{
            x,y,x+w,y,x+w,y+h,0,y+h
        })
    self.dirty = true
end

function lighter:addPolygon(verts)
    if clockwise(verts[1],verts[2],
        verts[3],verts[4],
        verts[5],verts[6]) then
        table.insert(self.polys,{unpack(verts)})
    else
        local verts_new = {}
        for i,v in ipairs(verts) do
            verts_new[#verts-i+1] = v 
        end
        table.insert(self.polys,verts_new)
    end
    self.dirty = true
end

function lighter:makeRays()
    for _,light in ipairs(self.lights) do
        light.rays = {}
        for _,poly in ipairs(self.polys) do
            local ox,oy = light.x,light.y
            for tx,ty in polygon(poly) do
                local dist = distance(ox,oy,tx,ty)
                local dir = math.atan2(tx-ox,ty-oy)
                table.insert(light.rays,{ox = ox,oy = oy,
                        tx = tx,ty = ty})
                table.insert(light.rays,{diff = light.diff,ox = ox,oy = oy,
                        tx = tx+dist*math.sin(dir+light.diff),ty = ty+dist*math.cos(dir+light.diff)})
                table.insert(light.rays,{diff = -light.diff,ox = ox,oy = oy,
                        tx = tx+dist*math.sin(dir-light.diff),ty = ty+dist*math.cos(dir-light.diff)})
            end
            
           
            local dir = light.dir - light.arc/2 --两端的光线
            local tx,ty = ox + math.sin(dir)*100,oy + math.cos(dir)*100
            table.insert(light.rays,{ox = ox,oy = oy,
                        tx = tx,ty = ty})
            local dir = light.dir + light.arc/2
            local tx,ty = ox + math.sin(dir)*100,oy + math.cos(dir)*100
            table.insert(light.rays,{ox = ox,oy = oy,
                        tx = tx,ty = ty})
        end
    end
    self.dirty = false
end

function lighter:cast()
    if self.dirty then
        self:makeRays()
    end
    
    for _,light in ipairs(self.lights) do
        for _,ray in ipairs(light.rays) do
            ray.ix = nil;ray.iy = nil;ray.dist = math.huge;ray.angle = nil
            for _, poly in ipairs(self.polys) do
                for ox,oy,i in polygon(poly) do
                    local tx,ty = poly[i*2+1] or poly[1],poly[i*2+2] or poly[2] 
                    local ix,iy,dist = intersection(
                        light.x,light.y,
                        ray.tx-light.x,ray.ty-light.y,
                        ox,oy,
                        tx-ox,ty-oy
                    )
                    if dist and ray.dist> dist then
                        ray.ix,ray.iy,ray.dist = ix,iy,dist
                        ray.angle = math.atan2(iy-light.y,ix-light.x)
                    end
                end
            end
        end
        
    end
end



function lighter:render()
    self:cast()
    love.graphics.setCanvas({self.canvas, stencil=true})
    love.graphics.clear(self.darkness)
    love.graphics.setBlendMode("replace")
    for _,light in ipairs(self.lights) do
        local function stencil()
            local rayInArc = {}
            for _,ray in ipairs(light.rays) do
                if ray.angle then
                    table.insert(rayInArc,ray)
                end
            end
            table.sort(rayInArc,function(a,b) return a.angle<b.angle end)
            for i,ray in ipairs(rayInArc) do
                local nr = rayInArc[i+1] or rayInArc[1]
                if not nr.ix or not nr.iy then break end
                love.graphics.polygon("fill",
                    light.x,light.y,ray.ix,ray.iy,
                    nr.ix,nr.iy)
            end
            love.graphics.arc( "fill", "pie", light.x, light.y, light.r, light.dir - light.arc/2, light.dir + light.arc/2 )
        end
        
        love.graphics.stencil(stencil, "increment", 1)
        love.graphics.setStencilTest("greater", 1)
        local r,g,b,a = unpack(light.color)
        love.graphics.setColor(r,g,b,a*light.power)
        love.graphics.draw(self.circle,light.x,light.y,0,light.r,light.r)
        love.graphics.setStencilTest()
    end
    love.graphics.setCanvas()
    love.graphics.setColor(1,1,1)
    love.graphics.draw(self.canvas)
    self:renderObj()
end

function lighter:renderObj()
    love.graphics.setColor(1,1,1)
    for _,poly in ipairs(self.polys) do
        love.graphics.polygon("line",poly)
    end
end

function lighter:debugDraw()
    if not debug then return end
    
    for _,light in ipairs(self.lights) do
        local rayInArc = {}
        for _,ray in ipairs(light.rays) do
            if ray.angle and clamp(ray.angle,light.dir-light.arc/2,light.dir+light.arc/2) == norm(ray.angle) 
                then
                table.insert(rayInArc,ray)
            end
        end
        table.sort(rayInArc,function(a,b) return a.angle<b.angle end)
        for i,ray in ipairs(rayInArc) do
            love.graphics.setColor(1,1,1)
            love.graphics.line(light.x,light.y,ray.ix,ray.iy)
            love.graphics.circle("fill",ray.ix,ray.iy,5)
            love.graphics.setColor(0.5,0.5,0.8,0.3)
            local nr = rayInArc[i+1] or rayInArc[1]
            if not nr.ix or not nr.iy then break end
            love.graphics.polygon("fill",
                light.x,light.y,ray.ix,ray.iy,
                nr.ix,nr.iy)
        end
    end
    
    love.graphics.setColor(1,1,1)
    for _,poly in ipairs(self.polys) do
        love.graphics.polygon("line",poly)
    end

end


return lighter